import os
import importlib.util
from pathlib import Path

import openai
import lancedb
import json

from mcp.server.fastmcp import FastMCP, Context
from pydantic import BaseModel, Field
from typing import Any, Optional, List, cast
from openai.types.chat import ChatCompletionSystemMessageParam


def register_u_tool(server: FastMCP, u_tool_open_ai_api_key: str):
    # LLM_MODEL = "gpt-4o"         # Input $2.5,   Output $10
    # LLM_MODEL = "gpt-4-turbo"    # Input $10.00, Output $30
    LLM_MODEL = "gpt-4o-mini"      # Input $0.15,  Output $0.60

    EMBEDDING_MODEL = "text-embedding-3-small"
    spec = importlib.util.find_spec("mcp_server_twelve_data")
    MODULE_PATH = Path(spec.origin).resolve()
    PACKAGE_ROOT = MODULE_PATH.parent  # src/mcp_server_twelve_data
    DB_PATH = str(PACKAGE_ROOT / "resources" / "endpoints.lancedb")
    TOP_N = 35

    class UToolResponse(BaseModel):
        """Response object returned by the u-tool."""

        top_candidates: List[str] = Field(
            ..., description="List of tool operationIds considered by the vector search."
        )
        selected_tool: Optional[str] = Field(
            None, description="Name (operationId) of the tool selected by the LLM."
        )
        param: Optional[dict] = Field(
            None, description="Parameters passed to the selected tool."
        )
        response: Optional[Any] = Field(
            None, description="Result returned by the selected tool."
        )
        error: Optional[str] = Field(
            None, description="Error message, if tool resolution or execution fails."
        )
        motivation: Optional[str] = Field(
            None, description="Brief explanation why the LLM chose this tool."
        )

    def build_openai_tools_subset(tool_list):
        tools = []
        for tool in tool_list:
            tools.append({
                "type": "function",
                "function": {
                    "name": tool.name,
                    "description": tool.description or "No description provided.",
                    "parameters": tool.parameters
                }
            })
        return tools

    os.environ["OPENAI_API_KEY"] = u_tool_open_ai_api_key
    all_tools = server._tool_manager._tools
    server._tool_manager._tools = {}  # leave only u-tool

    db = lancedb.connect(DB_PATH)
    table = db.open_table("endpoints")

    @server.tool(name="u-tool")
    async def u_tool(query: str, ctx: Context) -> UToolResponse:
        """
        A universal tool router for the MCP system, designed for the Twelve Data API.

        This tool accepts a natural language query in English and performs the following:
        1. Uses vector search to retrieve the top-N relevant Twelve Data endpoints.
        2. Sends the query and tool descriptions to OpenAI's gpt-4o with function calling.
        3. The model selects the most appropriate tool and generates the input parameters.
        4. The selected endpoint (tool) is executed and its response is returned.

        Supported endpoint categories (from Twelve Data docs):
        - Market & Reference: price, quote, symbol_search, stocks, exchanges, market_state
        - Time Series: time_series, eod, splits, dividends, etc.
        - Technical Indicators: rsi, macd, ema, bbands, atr, vwap, and 100+ others
        - Fundamentals & Reports: earnings, earnings_estimate, income_statement,
          balance_sheet, cash_flow, statistics, profile, ipo_calendar, analyst_ratings
        - Currency & Crypto: currency_conversion, exchange_rate, price_target
        - Mutual Funds / ETFs: funds, mutual_funds/type, mutual_funds/world
        - Misc Utilities: logo, calendar endpoints, time_series_calendar, etc.
        """
        client = openai.OpenAI()
        candidate_ids: List[str]

        try:
            embedding = client.embeddings.create(
                model=EMBEDDING_MODEL,
                input=[query]
            ).data[0].embedding

            results = table.search(embedding).metric("cosine").limit(TOP_N).to_list()  # type: ignore[attr-defined]
            candidate_ids = [r["id"] for r in results]

        except Exception as e:
            return UToolResponse(
                top_candidates=[],
                selected_tool=None,
                param=None,
                response=None,
                error=f"Embedding or vector search failed: {e}",
                motivation=None,
            )

        filtered_tools = [tool for tool in all_tools.values() if tool.name in candidate_ids]  # type: ignore
        openai_tools = build_openai_tools_subset(filtered_tools)

        prompt = (
            "You are a function-calling assistant. Based on the user query, "
            "you must select the most appropriate function from the provided tools and return "
            "a valid tool call with all required parameters. "
            "Before the function call, provide a brief plain-text explanation (1–2 sentences) of "
            "why you chose that function, based on the user's intent and tool descriptions."
        )

        try:
            response = client.chat.completions.create(
                model=LLM_MODEL,
                messages=[
                    cast(ChatCompletionSystemMessageParam, {"role": "system", "content": prompt}),
                    cast(ChatCompletionSystemMessageParam, {"role": "user", "content": query})
                ],
                tools=openai_tools,
                tool_choice="required",
                temperature=0,
            )

            call = response.choices[0].message.tool_calls[0]
            name = call.function.name
            arguments = json.loads(call.function.arguments)
            # all tools require single parameter with nested attributes, but sometimes LLM flattens it
            if "params" not in arguments:
                arguments = {"params": arguments}
            choice = response.choices[0].message
            motivation_text = choice.content.strip() if choice.content else None

        except Exception as e:
            return UToolResponse(
                top_candidates=candidate_ids,
                selected_tool=None,
                param=None,
                response=None,
                error=f"LLM did not return valid tool call: {e}",
                motivation=None,
            )

        tool = all_tools.get(name)
        if not tool:
            return UToolResponse(
                top_candidates=candidate_ids,
                selected_tool=name,
                param=arguments,
                response=None,
                error=f"Tool '{name}' not found in MCP",
                motivation=None,
            )

        try:
            params_type = tool.fn_metadata.arg_model.model_fields["params"].annotation
            arguments['params'] = params_type(**arguments['params'])
            arguments['ctx'] = ctx

            result = await tool.fn(**arguments)
            return UToolResponse(
                top_candidates=candidate_ids,
                selected_tool=name,
                param=arguments,
                response=result,
                error=None,
                motivation=motivation_text,
            )
        except Exception as e:
            return UToolResponse(
                top_candidates=candidate_ids,
                selected_tool=name,
                param=arguments,
                response=None,
                error=str(e),
                motivation=None,
            )



import os
import lancedb
import openai

# Константы
EMBEDDING_MODEL = "text-embedding-3-small"

def generate_embedding(text: str) -> list[float]:
    client = openai.OpenAI()
    response = client.embeddings.create(
        model=EMBEDDING_MODEL,
        input=[text]
    )
    return response.data[0].embedding

def evaluate_queries(query_target_pairs: list[tuple[str, str]]):
    db_path = os.getenv('LANCEDB_PATH', '../extra/endpoints.lancedb')
    db = lancedb.connect(db_path)
    table = db.open_table("endpoints")

    results_summary = []

    for user_query, target_endpoint in query_target_pairs:
        query_vec = generate_embedding(user_query)
        results = table.search(query_vec).metric("cosine").limit(30).to_list()

        top_endpoint = results[0]['path'] if results else None
        target_position = next((i for i, r in enumerate(results) if r['path'] == target_endpoint), None)

        print(f"Query: {user_query}")
        print(f"Target endpoint: {target_endpoint}")
        print(f"Top 1 endpoint: {top_endpoint}")
        print(f"Target position in top 30: {target_position}\n")

        results_summary.append((user_query, target_endpoint, top_endpoint, target_position))

    return results_summary


def main():
    query_target_pairs = [
        ("Show me intraday stock prices for Tesla (TSLA) with 1-minute intervals for the past 3 hours.", "/time_series"),
        ("What is the current exchange rate between USD and EUR?", "/price"),
        ("Get the RSI indicator for Apple (AAPL) over the last 14 days.", "/rsi"),
        ("When did Amazon last split its stock?", "/splits"),
        ("Give me daily closing prices for Bitcoin in the past 6 months.", "/time_series"),
        ("Show the MACD for Microsoft.", "/macd"),
        ("Get Google earnings reports for the last year.", "/earnings"),
        ("Fetch dividend history for Johnson & Johnson.", "/dividends"),
        ("Give me fundamentals for Netflix including P/E ratio.", "/fundamentals"),
        ("What is the latest stock quote for Nvidia?", "/quote"),
        ("Retrieve the Bollinger Bands for Apple.", "/bbands"),
        ("What is the VWAP for Tesla?", "/vwap"),
        ("Get ATR indicator for Amazon.", "/atr"),
        ("What is the stochastic oscillator for MSFT?", "/stoch"),
        ("Show me the EMA for S&P 500.", "/ema"),
        ("Retrieve the ADX indicator for crude oil.", "/adx"),
        ("Get the OBV for Bitcoin.", "/obv"),
        ("What is the highest stock price of Apple in the last 30 days?", "/max"),
        ("Give me the minimum price for TSLA in January 2024.", "/min"),
        ("Get the ROC indicator for Ethereum.", "/roc"),
    ]

    results = evaluate_queries(query_target_pairs)

    print("\nSummary:")
    for row in results:
        print(row)


if __name__ == "__main__":
    main()



import json
from pathlib import Path
import csv

OPENAPI_PATH = "../extra/openapi_clean.json"
ENDPOINTS_PATH = "../extra/endpoints_spec_en.csv"
OUTPUT_PATH = "../data/tools_autogen.py"


def load_endpoint_list(path):
    with open(path, newline='', encoding='utf-8') as csvfile:
        reader = csv.reader(csvfile)
        next(reader, None)  # skip header
        return [row[0] for row in reader]


def generate_code(endpoints, spec):
    ops = []
    for path in endpoints:
        if path not in spec.get('paths', {}):
            continue
        for method, details in spec['paths'][path].items():
            op_id = details.get('operationId')
            description = details.get('description', '').strip().replace('"', '\\"').replace('\n', ' ')
            if not op_id:
                continue
            endpoint_key = path.lstrip('/')
            ops.append((op_id, description, endpoint_key))

    lines = []
    # Imports
    lines.append('from mcp.server import FastMCP')
    lines.append('from mcp.server.fastmcp import Context')
    lines.append('')

    # Request model imports
    for op_id, _, _ in ops:
        lines.append(f'from .request_models import {op_id}Request')
    lines.append('')

    # Response model imports
    for op_id, _, _ in ops:
        lines.append(f'from .response_models import {op_id}200Response')
    lines.append('')

    # register_all_tools function
    lines.append('def register_all_tools(server: FastMCP, _call_endpoint):')
    for op_id, description, endpoint_key in ops:
        lines.append(f'    @server.tool(name="{op_id}",')
        lines.append(f'                 description="{description}")')
        lines.append(f'    async def {op_id}(params: {op_id}Request, ctx: Context) -> {op_id}200Response:')
        lines.append(f'        return await _call_endpoint("{endpoint_key}", params, {op_id}200Response, ctx)')
        lines.append('')

    return "\n".join(lines)


def main():
    endpoints = load_endpoint_list(ENDPOINTS_PATH)
    with open(OPENAPI_PATH, encoding='utf-8') as f:
        spec = json.load(f)

    code = generate_code(endpoints, spec)
    header = '# AUTOGENERATED FILE - DO NOT EDIT MANUALLY' + '\n\n'
    Path(OUTPUT_PATH).write_text(header + code, encoding='utf-8')


if __name__ == '__main__':
    main()



import os
import yaml
import json
import re
from collections import defaultdict

input_path = "../data/openapi01.06.2025.json"
output_dir = "../data"

mirrors = [
    "https://api-reference-data.twelvedata.com",
    "https://api-time-series.twelvedata.com",
    "https://api-mutual-funds.twelvedata.com",
    "https://api-etfs.twelvedata.com",
    "https://api-fundamental.twelvedata.com",
    "https://api-analysis.twelvedata.com",
    "https://api-regulator.twelvedata.com",
    "https://api-ti-overlap-studies.twelvedata.com",
    "https://api-ti-volume-indicators.twelvedata.com",
    "https://api-ti-price-transform.twelvedata.com",
    "https://api-ti-cycle-indicators.twelvedata.com",
    "https://api-ti-statistics-functions.twelvedata.com"
]

allowed_endpoints = [
    # Reference Data
    "/stocks",
    "/forex_pairs",
    "/cryptocurrencies",
    "/funds",
    "/bonds",
    "/etfs",
    "/commodities",
    "/cross_listings",
    "/exchanges",
    "/exchange_schedule",
    "/cryptocurrency_exchanges",
    "/market_state",
    "/instrument_type",
    "/countries",
    "/earliest_timestamp",
    "/symbol_search",
    "/intervals",

    # Core Data
    "/time_series",
    "/time_series/cross",
    "/exchange_rate",
    "/currency_conversion",
    "/quote",
    "/price",
    "/eod",
    "/market_movers/{market}",

    # Mutual Funds
    "/mutual_funds/list",
    "/mutual_funds/family",
    "/mutual_funds/type",
    "/mutual_funds/world",
    "/mutual_funds/world/summary",
    "/mutual_funds/world/performance",
    "/mutual_funds/world/risk",
    "/mutual_funds/world/ratings",
    "/mutual_funds/world/composition",
    "/mutual_funds/world/purchase_info",
    "/mutual_funds/world/sustainability",

    # ETFs
    "/etfs/list",
    "/etfs/family",
    "/etfs/type",
    "/etfs/world",
    "/etfs/world/summary",
    "/etfs/world/performance",
    "/etfs/world/risk",
    "/etfs/world/composition",

    # Fundamentals
    "/balance_sheet",
    "/balance_sheet/consolidated",
    "/cash_flow",
    "/cash_flow/consolidated",
    "/dividends",
    "/dividends_calendar",
    "/earnings",
    "/income_statement",
    "/income_statement/consolidated",
    "/ipo_calendar",
    "/key_executives",
    "/last_change/{endpoint}",
    "/logo",
    "/market_cap",
    "/profile",
    "/splits",
    "/splits_calendar",
    "/statistics",

    # Analysis
    "/analyst_ratings/light",
    "/analyst_ratings/us_equities",
    "/earnings_estimate",
    "/revenue_estimate",
    "/eps_trend",
    "/eps_revisions",
    "/growth_estimates",
    "/price_target",
    "/recommendations",
    "/earnings_calendar",

    # Regulatory
    "/tax_info",
    "/edgar_filings/archive",
    "/insider_transactions",
    "/direct_holders",
    "/fund_holders",
    "/institutional_holders",
    "/sanctions/{source}",
]

added_endpoints = []


def load_spec(path):
    with open(path, 'r', encoding='utf-8') as f:
        if path.lower().endswith(('.yaml', '.yml')):
            return yaml.safe_load(f)
        return json.load(f)


def dump_spec(spec, path):
    with open(path, 'w', encoding='utf-8') as f:
        if path.lower().endswith(('.yaml', '.yml')):
            yaml.safe_dump(spec, f, sort_keys=False, allow_unicode=True)
        else:
            json.dump(spec, f, ensure_ascii=False, indent=2)


def find_refs(obj):
    refs = set()
    if isinstance(obj, dict):
        for k, v in obj.items():
            if k == '$ref' and isinstance(v, str):
                refs.add(v)
            else:
                refs |= find_refs(v)
    elif isinstance(obj, list):
        for item in obj:
            refs |= find_refs(item)
    return refs


def split_paths(keys, chunk_size=25):
    for i in range(0, len(keys), chunk_size):
        yield keys[i:i + chunk_size]


def filter_paths(all_paths, allowed_list):
    """
    Returns a list of only those paths from all_paths that are present in allowed_list.
    Additionally, prints paths that are in allowed_list but not found in all_paths.
    """
    f_paths = [path for path in all_paths if path in allowed_list]
    missing_paths = [path for path in allowed_list if path not in f_paths]
    if missing_paths:
        print("Paths in allowed_list but not found in all_paths:", missing_paths)
    return f_paths


def prune_components(full_components, used_refs):
    pattern = re.compile(r'^#/components/([^/]+)/(.+)$')
    used = defaultdict(set)

    # Mark direct references
    for ref in used_refs:
        m = pattern.match(ref)
        if m:
            comp_type, comp_name = m.group(1), m.group(2)
            used[comp_type].add(comp_name)

    # Recursively include nested references
    changed = True
    while changed:
        changed = False
        for comp_type, names in list(used.items()):
            for name in list(names):
                definition = full_components.get(comp_type, {}).get(name)
                if definition:
                    for r in find_refs(definition):
                        m2 = pattern.match(r)
                        if m2:
                            ct, cn = m2.group(1), m2.group(2)
                            if cn not in used[ct]:
                                used[ct].add(cn)
                                changed = True

    # Assemble a limited set
    pruned = {}
    for comp_type, defs in full_components.items():
        if comp_type in used:
            kept = {n: defs[n] for n in defs if n in used[comp_type]}
            if kept:
                pruned[comp_type] = kept
    return pruned


def trim_fields(obj):
    """
    Recursively trims string fields:
    - 'description' to 300 characters
    - 'example' to 700 characters
    """
    if isinstance(obj, dict):
        for k, v in obj.items():
            if k == "description" and isinstance(v, str):
                if len(v) > 300:
                    obj[k] = v[:300]
            elif k == "example" and isinstance(v, str):
                if len(v) > 700:
                    obj[k] = v[:700]
            else:
                trim_fields(v)
    elif isinstance(obj, list):
        for item in obj:
            trim_fields(item)


def add_empty_properties(obj):
    """
    Recursively searches for all occurrences of the 'schema' key.
    If its value is a dict and this dict does not have 'properties',
    adds 'properties': {}.
    """
    if isinstance(obj, dict):
        for k, v in obj.items():
            if k == "schema" and isinstance(v, dict):
                if "properties" not in v:
                    v["properties"] = {}
                # Continue deep traversal within the found schema
                add_empty_properties(v)
            else:
                add_empty_properties(v)
    elif isinstance(obj, list):
        for item in obj:
            add_empty_properties(item)


def main():
    os.makedirs(output_dir, exist_ok=True)

    spec = load_spec(input_path)
    all_paths = list(spec.get('paths', {}).keys())
    f_paths = filter_paths(all_paths, allowed_endpoints)
    chunks = list(split_paths(f_paths))

    for idx, paths_chunk in enumerate(chunks, start=1):
        # Build a new part of the specification
        new_spec = {
            'openapi': spec.get('openapi'),
            'info': spec.get('info'),
            'servers': (
                [{'url': mirrors[idx - 1]}]
                if idx - 1 < len(mirrors)
                else spec.get('servers', [])
            ),
            'paths': {k: spec['paths'][k] for k in paths_chunk}
        }

        # Trim long fields and add missing 'properties'
        add_empty_properties(new_spec)
        trim_fields(new_spec)

        # Prune components
        used_refs = find_refs(new_spec['paths'])
        pruned = prune_components(spec.get('components', {}), used_refs)
        if pruned:
            new_spec['components'] = pruned

        # Calculate metrics and save the file
        new_paths_count = len(new_spec['paths'])
        new_components_count = sum(
            len(v) for v in new_spec.get('components', {}).values()
        )
        out_file = os.path.join(
            output_dir,
            f"{os.path.splitext(os.path.basename(input_path))[0]}"
            f"_part{idx}{os.path.splitext(input_path)[1]}"
        )
        dump_spec(new_spec, out_file)
        print(
            f"Part {idx}: {new_paths_count} paths, "
            f"{new_components_count} components -> {out_file}"
        )


if __name__ == "__main__":
    main()



from datamodel_code_generator import InputFileType, generate, PythonVersion
from pathlib import Path

openapi_path = '../extra/openapi_clean.json'   # путь к спецификации
output_path = '../data/response_models.py'        # куда сохранить

# Чтение исходного файла
with open(openapi_path, 'r', encoding='utf-8') as f:
    openapi_content = f.read()

# Генерация моделей
generate(
    input_=openapi_content,
    input_file_type=InputFileType.OpenAPI,
    output=Path(output_path),
    class_name=None,
    target_python_version=PythonVersion.PY_313,
)

print(f"Генерация завершена: {output_path}")
